/* Copyright (c) 2013, David A. Clunie DBA Pixelmed Publishing. All rights reserved. */

package com.pixelmed.ucum;

import antlr.*;
import antlr.collections.*;

import java.io.InputStream;
import java.io.StringBufferInputStream;

/**
 * <p>A class of static methods to convert UCUM code strings into a canonical form
 * and to determine a conversion factor when their canonical forms are equivalent
 * apart from a constant multiplier value.</p>
 *
 * <p>The canonical form used here is defined to be that involving use of the base units only, or
 * those units that cannot be converted to base units (typically those defined as meaning
 * unity, such as Hounsfield Units for CT, etc.), sorted into a canonical order, with
 * exponents fully expanded, prefixes expanded into constans, constants folded, and from which redundant terms in the
 * numerator and denominator have been eliminated.</p>
 *
 * <p>Note that the class does not attempt to check the appropriateness of a conversion; the user
 * is responsible for establishing that the two units to be converted are both of the same type
 * of quantity. For example, both Herz (units of frequency) and Becquerel (units of radioactivity)
 * are defined with the same canonical form (s-1), though one represents the number of cycles
 * per second and the other the number of counts per second.</p>
 *
 * <p>Also, a handful of units cannot be converted because the conversion is not based on a
 * linear multiplication. This includes the units of pH, the measures of level, which are
 * logarithmic (e.g. Bel), and most notably, units of temperature. For the purpose of
 * converting temperature differences, three experimental extensions to UCUM 1.4 are
 * provided, [degK_delta], [degC_delta] and [degF_delta], which can be converted by
 * a constant multiplier. These are useful, for example, in meteorology to convert the
 * atmospheric lapse rate (change in temperature per unit of altitude) from one unit
 * to another, without concern for the absolute temperature.</p>
 *
 * <p>A {@link #main(String[]) main()} method is provided for testing and the convenience of command line execution.</p>
 *
 * <p>For example, the following command extracts the conversion factor from centimetres per second to US furlongs per hour:</p>
 *
 * <pre>
 *	% java -cp ./pixelmed_ucum.jar:./antlr-2.7.5.jar com.pixelmed.ucum.UCUMAnalyzer "cm/s" "[fur_us]/h"
 *	Conversion factor = 5.588011176022351
 * </pre>
 *
 * <p>To illustrate the canonical forms used in the above examples, the following commands generate LISP-like output describing the constant, the numerator and the denominator:</p>
 *
 * <pre>
 *	% java -cp ./pixelmed_ucum.jar:./antlr-2.7.5.jar com.pixelmed.ucum.UCUMAnalyzer "cm/s"
 *	Canonical form = ( / ( . 0.01  m ) ( .  s ) )
 * </pre>
 *
 * <pre>
 *	% java -cp ./pixelmed_ucum.jar:./antlr-2.7.5.jar com.pixelmed.ucum.UCUMAnalyzer "[fur_us]/h"
 *	Canonical form = ( / ( . 0.05588011176022351  m ) ( .  s ) )
 * </pre>
 *
 * @author	dclunie
 */

public class UCUMAnalyzer {

	private static ASTFactory astFactory = new ASTFactory();
	
	/**
	 * <p>Sort an AST by integer numeric value of token type.</p>
	 *
	 * @param	ast				an abstract syntax tree to sort
	 * @return					a sorted abstract syntax tree containing duplicate nodes
	 */
	private static AST sortByTokenType(AST src) {
		if (src != null && src.getNextSibling() != null) {
			boolean swapped = true;
			while (swapped) {
//System.out.println("UCUMAnalyzer.sortByTokenType(): list = "+src.toStringList());
				swapped = false;
				AST dst = null;
				AST last = null;
				while (src != null) {
//System.out.println("UCUMAnalyzer.sortByTokenType(): src = "+src.getText());
					AST first = null;
					AST second = null;
					AST next = src.getNextSibling();
//System.out.println("UCUMAnalyzer.sortByTokenType(): next = "+(next == null ? "null" : next.getText()));
					if (next != null && src.getType() > next.getType()) {
						first  = astFactory.dup(next);
						second = astFactory.dup(src);
//System.out.println("UCUMAnalyzer.sortByTokenType(): swapping, first = "+first.toStringList());
//System.out.println("UCUMAnalyzer.sortByTokenType(): swapping, second = "+second.toStringList());
						swapped = true;
						src = next.getNextSibling();
					}
					else {
						first  = astFactory.dup(src);
						second = null;
//System.out.println("UCUMAnalyzer.sortByTokenType(): copying, first = "+first.toStringList());
						src = next;
					}
					if (dst == null) {
						dst = first;
//System.out.println("UCUMAnalyzer.sortByTokenType(): starting new list, dst = "+dst.toStringList());
						last=first;
					}
					else {
						last.setNextSibling(first);
//System.out.println("UCUMAnalyzer.sortByTokenType(): adding first, dst now = "+dst.toStringList());
						last=first;
					}
					if (second != null) {
						last.setNextSibling(second);
//System.out.println("UCUMAnalyzer.sortByTokenType(): adding second, dst now = "+dst.toStringList());
						last=second;
					}
				}
				src = dst;
			}
		}
//System.out.println("UCUMAnalyzer.sortByTokenType(): returning = "+src.toStringList());
		return src;
	}
	
	/**
	 * <p>Factor terms common to the numerator and the denominator and fold constants.</p>
	 *
	 * @param	ast				an abstract syntax tree that has already been flattened (is no longer binary)
	 * @param	debugMode			values greater than zero activate debugging messages
	 * @return					the canonical form of a UCUM code
	 * @exception	UCUMConversionException		if the UCUM code cannot be cannot be analyzed
	 */
	private static UCUMCanonicalStructure factorOutCommonTerms(AST src,int debugMode) throws UCUMConversionException {
		// by this time, the source tree should be of the form:
		//
		// ( / ( . constant numeratorTree ) denominatorTree )
		//
		// or
		//
		// ( / ( . constant numeratorTree ) ( . denominatorTree ) )
		//
		// or
		//
		// ( . constant numeratorTree )
		//
		// or
		//
		// ( constant )
		//
		// but neither the numeratorTree and denominatorTree are necessarily sorted by token order
		
		double numeratorConstantValue = 1;
		double denominatorConstantValue = 1;
		AST newNumerator = null;
		AST newDenominator = null;
		
		if (src == null) {
			return null;
		}
		
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): src = "+src.toStringList());
		int srcType = src.getType();
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): srcType = "+srcType);

		if (srcType == UCUMRewriterTokenTypes.DIGITS && src.getFirstChild() == null && src.getNextSibling() == null) {
			numeratorConstantValue = Double.parseDouble(src.getText());
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): got constant only, value = "+numeratorConstantValue);
		}
		else if (srcType == UCUMRewriterTokenTypes.MULTIPLY) {
			AST numerator = src.getFirstChild();
			if (numerator != null) {
				numerator = sortByTokenType(numerator);
			}
			while (numerator != null && numerator.getType() == UCUMRewriterTokenTypes.DIGITS) {
				numeratorConstantValue *= Double.parseDouble(numerator.getText());
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): constant numerator term now = "+numeratorConstantValue);
				numerator = numerator.getNextSibling();
			}
			newNumerator = numerator;
		}
		else if (srcType == UCUMRewriterTokenTypes.DIVIDE) {
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): root is divide");
			AST numerator = src.getFirstChild();
			if (numerator != null) {
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): got numerator");
				AST denominator = numerator.getNextSibling();
				if (denominator != null) {
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): got denominator = "+denominator.toStringList());
					if (denominator.getNextSibling() == null) {
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): no further children of root as expected");
						if (numerator.getType() == UCUMRewriterTokenTypes.MULTIPLY) {
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): numerator root is multiply");
							numerator = numerator.getFirstChild();
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): numerator now");
						}
						else if (numerator.getType() == UCUMRewriterTokenTypes.DIGITS) {
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): numerator is constant");
							numerator.setNextSibling(null);		// disconnect from denominator of divide that follows
						}
						else {
							throw new UCUMConversionException("Unexpected form of numerator whilst factoring out common terms");
						}
						
						if (numerator != null) {
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): got numerator = "+numerator.toStringList());
							if (denominator.getType() == UCUMRewriterTokenTypes.MULTIPLY) {
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): denominator root is multiply");
								denominator = denominator.getFirstChild();
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): got denominator = "+denominator.toStringList());
							}
									
							// first need to sort the numerator and denominator
							numerator = sortByTokenType(numerator);
							denominator = sortByTokenType(denominator);

							AST newNumeratorLast = null;
							AST newDenominatorLast = null;

							// walk numerator and denominator, removing common terms
							do {
								int numeratorType = numerator == null ? 99999 : numerator.getType();
								int denominatorType = denominator == null ? 99999 : denominator.getType();
								if (numeratorType == denominatorType && numeratorType != 99999) {
									if (numeratorType == UCUMRewriterTokenTypes.DIGITS) {
										numeratorConstantValue *= Double.parseDouble(numerator.getText());
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): constant numerator term now = "+numeratorConstantValue);
										denominatorConstantValue *= Double.parseDouble(denominator.getText());
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): constant denominator term now = "+denominatorConstantValue);
									}
									else {
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): removing common factor term = "+numerator.getText());
									}
									numerator = numerator.getNextSibling();
									denominator = denominator.getNextSibling();
								}
								else if (numeratorType < denominatorType) {
									if (numeratorType != UCUMRewriterTokenTypes.DIGITS) {
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): keeping numerator term = "+numerator.getText());
										AST newNumeratorTerm = astFactory.dup(numerator);
										if (newNumerator == null) {
											newNumerator=newNumeratorTerm;
											newNumeratorLast=newNumeratorTerm;
										}
										else {
											newNumeratorLast.setNextSibling(newNumeratorTerm);
											newNumeratorLast=newNumeratorTerm;
										}
									}
									else {
										numeratorConstantValue *= Double.parseDouble(numerator.getText());
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): constant numerator term now = "+numeratorConstantValue);
									}
									numerator = numerator.getNextSibling();
								}
								else if (numeratorType > denominatorType)  {
									if (denominatorType != UCUMRewriterTokenTypes.DIGITS) {
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): keeping denominator term = "+denominator.getText());
										AST newDenominatorTerm = astFactory.dup(denominator);
										if (newDenominator == null) {
											newDenominator=newDenominatorTerm;
											newDenominatorLast=newDenominatorTerm;
										}
										else {
											newDenominatorLast.setNextSibling(newDenominatorTerm);
											newDenominatorLast=newDenominatorTerm;
										}
									}
									else {
										denominatorConstantValue *= Double.parseDouble(denominator.getText());
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): constant denominator term now = "+denominatorConstantValue);
									}
									denominator = denominator.getNextSibling();
								}
								// else both are 99999 and we are done
							} while (numerator != null || denominator != null);

//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): new numerator = "+(newNumerator == null ? "null" : newNumerator.toStringList()));
//System.out.println("UCUMAnalyzer.factorOutCommonTerms(): new denominator = "+(newDenominator == null ? "null" : newDenominator.toStringList()));

						}
					}
				}
			}
		}
		else {
			throw new UCUMConversionException("Unexpected form of canonical tree whilst factoring out common terms");
		}
		
		return new UCUMCanonicalStructure(numeratorConstantValue/denominatorConstantValue,newNumerator,newDenominator);
	}
	
	/**
	 * <p>Get the canonical form of a UCUM code.</p>
	 *
	 * @param	stream				a stream containing a UCUM code terminated by a new line
	 * @param	debugMode			values greater than zero activate debugging messages
	 * @return					the canonical form of a UCUM code
	 * @exception	UCUMConversionException		if the UCUM code cannot be parsed or a canonical form cannot be generated
	 */
	private static UCUMCanonicalStructure getCanonicalForm(InputStream stream,int debugMode) throws UCUMConversionException {
		AST t = null;
		try {
			UCUMLexer lexer = new UCUMLexer(stream);
			lexer.setDebugMode(debugMode);
			UCUMParser parser = new UCUMParser(lexer);
			parser.setDebugMode(debugMode);
			parser.line();
			t = parser.getAST();
			if (debugMode > 0) {
				System.out.println("List: "+t.toStringList());
			}
		}
		catch (Exception e) {
			throw new UCUMConversionException(e+" whilst parsing expression");
		}
		
		UCUMRewriter rewriter = new UCUMRewriter();
		rewriter.setDebugMode(debugMode);
		rewriter.setMode(1);	// binary tree rewriting
		AST tm1 = null;
		while (true) {
			try {
				rewriter.expr(t);
			}
			catch (Exception e) {
				throw new UCUMConversionException(e+" whilst rewriting "+ t.toStringList());
			}
			AST t2 = (CommonAST)rewriter.getAST();
			if (t.equalsTree(t2)) {
				break;		// keep rewriting tree until no change
			}
			else if (tm1 != null && t2.equalsTree(tm1)) {
				throw new UCUMConversionException("Infinite loop alternating between rewriting "+t.toStringList()+" to "+t2.toStringList());
			}
			tm1 = t;
			t = t2;
			if (debugMode > 1) {
				System.out.println("Rewriting list: "+t.toStringList());
			}
		}
		if (debugMode > 0) {
			System.out.println("Rewritten list: "+t.toStringList());
		}
		rewriter.setMode(2);	// binary tree rewriting
		try {
			rewriter.expr(t);
		}
		catch (Exception e) {
			throw new UCUMConversionException(e+" whilst flattening "+ t.toStringList());
		}
		AST t2 = (CommonAST)rewriter.getAST();
		t = t2;

		if (debugMode > 0) {
			System.out.println("Flattened list: "+t.toStringList());
		}
		return factorOutCommonTerms(t,debugMode);
	}

	/**
	 * <p>Get the canonical form of a UCUM code.</p>
	 *
	 * @param	s				the UCUM code
	 * @param	debugMode			values greater than zero activate debugging messages
	 * @return					the canonical form of a UCUM code
	 * @exception	UCUMConversionException		if the UCUM code cannot be parsed or a canonical form cannot be generated
	 */
	public static UCUMCanonicalStructure getCanonicalForm(String s,int debugMode) throws UCUMConversionException {
		return getCanonicalForm(new StringBufferInputStream(s+"\n"),debugMode);
	}

	/**
	 * <p>Attempt to convert one UCUM code into another if their canonical forms are equivalent other than a constant multiplier.</p>
	 *
	 * @param	c1				the canonical form of a UCUM code
	 * @param	c2				the canonical form of a UCUM code
	 * @return					the conversion factor by which to multiply values in c1 units to obtain values in c2 units
	 * @exception	UCUMConversionException		if the UCUM codes cannot be parsed or their canonical forms are not equivalent
	 */
	public static double getConversionFactor(UCUMCanonicalStructure c1,UCUMCanonicalStructure c2) throws UCUMConversionException {
		double conversion = 0;
		if (c1.equalsApartFromConstant(c2)) {
			conversion = c2.getConstant()/c1.getConstant();
		}
		else {
			throw new UCUMConversionException("Not units of the same type of quantity - cannot convert");
		}
		return conversion;
	}

	/**
	 * <p>Attempt to convert one UCUM code into another if their canonical forms are equivalent other than a constant multiplier.</p>
	 *
	 * @param	stream1				a stream containing a UCUM code terminated by a new line
	 * @param	stream2				a stream containing a UCUM code terminated by a new line
	 * @param	debugMode			values greater than zero activate debugging messages
	 * @return					the conversion factor by which to multiply values in stream1 units to obtain values in stream2 units
	 * @exception	UCUMConversionException		if the UCUM codes cannot be parsed or their canonical forms are not equivalent
	 */
	private static double getConversionFactor(InputStream stream1,InputStream stream2,int debugMode) throws UCUMConversionException {
		UCUMCanonicalStructure c1 = getCanonicalForm(stream1,debugMode);
		if (debugMode > 0) {
			System.out.println("Canonical form of first expression  = "+c1);
		}
		UCUMCanonicalStructure c2 = getCanonicalForm(stream2,debugMode);
		if (debugMode > 0) {
			System.out.println("Canonical form of second expression = "+c2);
		}
		return getConversionFactor(c1,c2);
	}

	/**
	 * <p>Attempt to convert one UCUM code into another if their canonical forms are equivalent other than a constant multiplier.</p>
	 *
	 * @param	s1				a UCUM code
	 * @param	s2				a UCUM code
	 * @param	debugMode			values greater than zero activate debugging messages
	 * @return					the conversion factor by which to multiply values in s1 units to obtain values in s2 units
	 * @exception	UCUMConversionException		if the UCUM codes cannot be parsed or their canonical forms are not equivalent
	 */
	public static double getConversionFactor(String s1,String s2,int debugMode) throws UCUMConversionException {
		InputStream parserInput1 = new StringBufferInputStream(s1+"\n");
		InputStream parserInput2 = new StringBufferInputStream(s2+"\n");
		return getConversionFactor(parserInput1,parserInput2,debugMode);
	}

	/**
	 * <p>Read UCUM strings from command line arguments or standard input and print their canonical form or attempt to convert between them.</p>
	 *
	 * <p>If the first argument is "-v" or "-vv", activate debugging messages.</p>
	 *
	 * <p>If there are no (remaining) arguments, read a single UCUM string from the standard input and print its canonical form.</p>
	 *
	 * <p>If there is a single (remaining) argument, regard it as a single UCUM string and print its canonical form.</p>
	 *
	 * <p>If there are two (remaining) arguments, regard them as UCUM strings and compare their canonical forms for equivalence and print a conversion factor.</p>
	 *
	 * @param	arg	array of zero, one, two or three strings
	 */
	public static void main(String[] arg) throws Exception {
		int argCount = arg.length;
		int argStart = 0;
		
		int debugMode = 0;
		if (argCount > 0) {
			if (arg[0].equals("-v")) {
				--argCount; ++argStart;
				debugMode = 1;
			}
			else if (arg[0].equals("-vv")) {
				--argCount; ++argStart;
				debugMode = 2;
			}
		}	
		
		try {
			if (argCount == 0) {
				UCUMCanonicalStructure c = getCanonicalForm(System.in,debugMode);
				System.out.println("Canonical form = "+c);
			}
			else if (argCount == 1) {
				UCUMCanonicalStructure c = getCanonicalForm(arg[argStart],debugMode);
				System.out.println("Canonical form = "+c);
			}
			else if (argCount == 2) {
				System.out.println("Conversion factor = "+getConversionFactor(arg[argStart],arg[argStart+1],debugMode));
			}
		}
		catch (Exception e) {
			System.err.println(e);
		}
	}
}
